//
//  LoginPanelController.m
//  Perforce
//
//  Created by Adam Czubernat on 13.05.2013.
//  Copyright (c) 2013 Perforce Software, Inc. All rights reserved.
//

#import "LoginPanelController.h"
#import <SystemConfiguration/SystemConfiguration.h>

#import "P4Workspace.h"
#import "P4Defaults.h"
#import "P4WorkspaceDefaults.h"
#import "ErrorPanelController.h"

@interface LoginPanelController () {
	
	NSString *host;
	NSString *username;
	NSArray *servers;

	enum {
		panelTypeLogin,
		panelTypeConnection,
		panelTypeRelogin,
	} panelType;
	
	// Login
	__weak IBOutlet NSView *loginView;
	__weak IBOutlet NSTextField *usernameTextField;
	__weak IBOutlet NSTextField *passwordTextField;
	__weak IBOutlet NSButton *loginCancelButton;
	
	// Connection
	__weak IBOutlet NSView *connectionView;
	__weak IBOutlet NSComboBox *connectionComboBox;
	__weak IBOutlet NSTextField *connectionAddressField;
	__weak IBOutlet NSTextField *connectionUsernameField;
	__weak IBOutlet NSTextField *connectionPasswordField;
	__weak IBOutlet NSButton *connectionCancelButton;
	
	// Relogin
	__weak IBOutlet NSView *reloginView;
	__weak IBOutlet NSTextField *reloginPasswordField;
	
	// Loading
	__weak IBOutlet NSView *loadingView;
	__weak IBOutlet NSProgressIndicator *loadingIndicator;
	__weak IBOutlet NSTextField *loadingLabel;
	
	// SSO
	__weak IBOutlet NSButton *ssoCheckbox;
}

- (void)loadWorkspace:(NSString *)workspace;
- (void)didLoadWorkspace:(NSArray *)workspaceInfo;
- (NSString *)defaultWorkspaceName;
- (void)loadDefaultWorkspace;
- (void)createDefaultWorkspace;

- (void)failWithError:(NSError *)error;
- (void)loginWithCredentials:(P4Credentials *)credentials;
- (void)didLogin;

- (void)showLoginView;
- (void)showConnectionView;
- (void)showReloginView;

- (IBAction)cancelPressed:(id)sender;

- (IBAction)loginPressed:(id)sender;
- (IBAction)connectionDetailsPressed:(id)sender;
- (IBAction)comboBoxSelected:(id)sender;
- (IBAction)connectPressed:(id)sender;

- (IBAction)reloginPressed:(id)sender;
- (IBAction)disconnectPressed:(id)sender;

- (IBAction)ssoCheckboxPressed:(id)sender;
@end

@implementation LoginPanelController
@synthesize delegate, allowsLastWorkspace, allowsSSOLogin;

- (void)windowDidLoad {
	[super windowDidLoad];
	
	host = [P4Defaults sharedInstance].host;
	username = [P4Defaults sharedInstance].username;
#ifdef DEBUG
//	[addressTextField setStringValue:@"test-server.homelinux.org:16660"];
//	[addressTextField setStringValue:@"ec2-50-16-43-78.compute-1.amazonaws.com:1666"];
//	[usernameTextField setStringValue:@"a.czubernat"];
//	[passwordTextField setStringValue:@"p@ssw0rd"];
	[passwordTextField setStringValue:@"adam23"];
#endif
	
	if (panelType == panelTypeLogin) {
		if (host.length)
			[self showLoginView];
		else
			[self showConnectionView];
	} else if (panelType == panelTypeConnection) {
		[self showConnectionView];
	} else if (panelType == panelTypeRelogin) {
		[self showReloginView];
	}
}

#pragma mark - Public

+ (LoginPanelController *)connectionPanel {
	LoginPanelController *panel = [[LoginPanelController alloc] init];
	panel->panelType = panelTypeConnection;
	return panel;
}

+ (LoginPanelController *)loginPanel {
	LoginPanelController *panel = [[LoginPanelController alloc] init];
	panel->panelType = panelTypeLogin;
	return panel;
	
}

+ (LoginPanelController *)reloginPanel {
	LoginPanelController *panel = [[LoginPanelController alloc] init];
	panel->panelType = panelTypeRelogin;
	return panel;
}

#pragma mark - Private

- (void)loadWorkspace:(NSString *)workspace {
	[[P4Workspace sharedInstance]
	 setWorkspace:workspace
	 response:^(P4Operation *operation, NSArray *response) {
		 [loadingIndicator stopAnimation:nil];
		 if (operation.errors) {
			 PSLog(@"Couldn't use workspace %@ %@", workspace, operation.error);
			 [self didLogin];
		 } else {
			 PSLog(@"Connected to workspace: %@", workspace);
			 [self didLoadWorkspace:response];
		 }
	 }];
}

- (void)didLoadWorkspace:(NSArray *)workspaceInfo {
	PSLogStore(@"Workspace info", @"%@", workspaceInfo);
	NSString *workspace = [[workspaceInfo firstObject] objectForKey:@"Client"];

	// Save last workspace details
	[P4WorkspaceDefaults setWorkspace:workspace];
	
	if ([delegate respondsToSelector:@selector(loginPanelDidLoginWithWorkspace:)])
		[delegate loginPanelDidLoginWithWorkspace:workspace];
}

- (NSString *)defaultWorkspaceName {
	NSString *machine = (__bridge NSString *)SCDynamicStoreCopyLocalHostName(NULL);
	return [NSString stringWithFormat:@"piper.%@.%@", machine, username];
}

- (void)loadDefaultWorkspace {
	NSString *workspace = [self defaultWorkspaceName];
	
	[[P4Workspace sharedInstance]
	 setWorkspace:workspace
	 response:^(P4Operation *operation, NSArray *response) {
		 NSArray *unknownClient = [operation errorsWithCode:P4ErrorUnknownClient];
		 [operation ignoreErrors:unknownClient];
		 if (operation.errors) {
			 [self failWithError:operation.error];
		 } else if (unknownClient) {
			 if ([[NSAlert
				   alertWithMessageText:@"\nWorkspace not found"
				   defaultButton:@"Create"
				   alternateButton:@"No"
				   otherButton:nil
				   informativeTextWithFormat:@"Do you want to create a default workspace?"]
				  runModal] != NSAlertDefaultReturn) {
				 [self didLogin]; // Skip workspace loading
				 return;
			 }
			 // Create workspace
			 PSLog(@"Creating default workspace...");
			 [self createDefaultWorkspace];
		 } else {
			 PSLog(@"Connected to default workspace: %@", workspace);
			 [self didLoadWorkspace:response];
		 }
	 }];
}

- (void)createDefaultWorkspace {
	[loadingLabel setStringValue:@"Configuring workspace..."];
	
	NSString *workspace = [self defaultWorkspaceName];
	
	NSURL *publicURL = [[[NSFileManager defaultManager]
						 URLsForDirectory:NSSharedPublicDirectory
						 inDomains:NSUserDomainMask] firstObject];
	NSURL *rootURL = [publicURL URLByAppendingPathComponent:workspace isDirectory:YES];
	if (![rootURL checkResourceIsReachableAndReturnError:NULL]) {
		[[NSFileManager defaultManager]
		 createDirectoryAtURL:rootURL withIntermediateDirectories:NO
		 attributes:nil error:NULL];
	}
	
	NSDictionary *workspaceSpecs =
	@{ @"Client" :	workspace,
	   @"Owner" :	[[P4Workspace sharedInstance] username],
	   @"Description" : @"Created by Piper",
	   @"Root" :	rootURL.path,
	   @"Options" :	@"noallwrite clobber rmdir nocompress unlocked",
	   @"SubmitOptions" : @"revertunchanged",
	   @"LineEnd" : @"local",
	   };

	[[P4Workspace sharedInstance]
	 createWorkspace:workspaceSpecs
	 response:^(P4Operation *operation, NSArray *response) {
		 if (operation.errors) {
			 [self failWithError:operation.error];
			 return;
		 }
		 PSLog(@"Created default workspace %@", workspace);
		 [self loadWorkspace:workspace];
	 }];
}

- (void)failWithError:(NSError *)error {
	[loadingIndicator stopAnimation:nil];
	[self showLoginView];
	
	ErrorPanelController *alert = [[ErrorPanelController alloc] init];
	[alert setTitle:@"\nLogin failed"];
    if ([error.localizedDescription rangeOfString:@"Connect to server failed"].location != NSNotFound) {
        [alert setMessage:@"It seems like we lost connection to the server. Please ensure that you are on the network."];
    } else {
        [alert setMessage:error.localizedDescription];
    }
	[alert presentWithMainWindow];
}

- (void)loginWithCredentials:(P4Credentials *)credentials {

	[self setPresentedView:loadingView animated:YES];
	[loadingIndicator startAnimation:nil];
	[loadingLabel setStringValue:@"Logging in..."];
	
	// Logout if logged in
	if ([[P4Workspace sharedInstance] isLoggedIn]) {
		[[P4Workspace sharedInstance] logout:^(P4Operation *operation, NSArray *response) {
			[self loginWithCredentials:credentials];
		}];
		return;
	}
	
	PSLog(@"Logging in...");
	
	// Response block
	P4ResponseBlock_t loginResponseBlock = ^(P4Operation *operation, NSArray *response) {		
		if (operation.errors) {
			[self failWithError:operation.error];
			return;
		}
		// Connected
		PSLog(@"Connected");
		// Save last login details
		[P4Defaults sharedInstance].host = credentials.address;
		[P4Defaults sharedInstance].username = credentials.username;
		
		// Automatic workspace selection
		NSString *lastWorkspace = [[P4Defaults sharedInstance].users objectForKey:username];
		if (allowsLastWorkspace && lastWorkspace) {
			PSLog(@"Loading last workspace \"%@\"...", lastWorkspace);
			[self loadWorkspace:lastWorkspace];
		// Load default workspace
		} else {
			PSLog(@"Loading default workspace...");
			[self loadDefaultWorkspace];
		}
	};
	
	// Connect with specified host
	[[P4Workspace sharedInstance]
	 connectWithCredentials:credentials
	 response:^(P4Operation *operation, NSArray *response) {		
		 if (operation.errors) {
			 [self failWithError:operation.error];
			 return;
		 }

		 if (allowsSSOLogin) {
			 [[P4Workspace sharedInstance]
			  loginWithSSO:credentials
			  response:loginResponseBlock];
		 } else {
			 [[P4Workspace sharedInstance]
			  loginWithCredentials:credentials
			  response:loginResponseBlock];
		 }
	 }];
}

- (void)didLogin {
	[loadingIndicator stopAnimation:nil];
	[self showLoginView];
	if ([delegate respondsToSelector:@selector(loginPanelDidLogin)])
		[delegate loginPanelDidLogin];
}

- (void)showLoginView {
	
	[loadingIndicator stopAnimation:nil];
	usernameTextField.stringValue = username ?: @"";
	[loginCancelButton setEnabled:
	 [[P4Workspace sharedInstance] isLoggedIn] &&
	 [[P4Workspace sharedInstance] workspace]];
		
	[self setPresentedView:loginView animated:YES];
}

- (void)showConnectionView {

	connectionAddressField.stringValue = host ?: @"";
	connectionUsernameField.stringValue = username ?: @"";
	[connectionCancelButton setEnabled:
	 [[P4Workspace sharedInstance] isLoggedIn] &&
	 [[P4Workspace sharedInstance] workspace]];
	
	// Load servers list
	NSString *plist = [[NSBundle mainBundle] pathForResource:@"Servers" ofType:@"plist"];
	servers = [NSArray arrayWithContentsOfFile:plist];
	NSArray *names = [servers valueForKeyPath:@"name"];
	NSArray *addresses = [servers valueForKeyPath:@"address"];
	
	[connectionComboBox removeAllItems];
	[connectionComboBox addItemsWithObjectValues:names];
	
	// Select server if last address matches one in the list
	NSInteger idx = [addresses indexOfObject:host];
	if (idx != NSNotFound)
		[connectionComboBox selectItemAtIndex:idx];
	
	[ssoCheckbox setState:allowsSSOLogin];
	
	[self setPresentedView:connectionView animated:YES];
	[self.window makeFirstResponder:connectionPasswordField];
}

- (void)showReloginView {
	[self setPresentedView:reloginView animated:YES];
}

#pragma mark - Actions

- (IBAction)cancelPressed:(id)sender {
	if ([delegate respondsToSelector:@selector(loginPanelDidCancel)])
		[delegate loginPanelDidCancel];
}

- (IBAction)loginPressed:(NSButton *)sender {
	
	P4Credentials *credentials = [[P4Credentials alloc] init];
	credentials.address = host;
	credentials.username = username = usernameTextField.stringValue;
	credentials.password = passwordTextField.stringValue;
	connectionPasswordField.stringValue = passwordTextField.stringValue;
	
	[self loginWithCredentials:credentials];
}

- (IBAction)connectionDetailsPressed:(id)sender {

	username = usernameTextField.stringValue;
	connectionPasswordField.stringValue = passwordTextField.stringValue;
	
	[self showConnectionView];
}

- (IBAction)comboBoxSelected:(id)sender {
	NSInteger idx = [connectionComboBox indexOfSelectedItem];
	NSString *address = [[servers objectAtIndex:idx] objectForKey:@"address"];
	connectionAddressField.stringValue = address ?: @"";
}

- (IBAction)connectPressed:(id)sender {

	P4Credentials *credentials = [[P4Credentials alloc] init];
	credentials.address = host = connectionAddressField.stringValue;
	credentials.username = username = connectionUsernameField.stringValue;
	credentials.password = connectionPasswordField.stringValue;
	passwordTextField.stringValue = connectionPasswordField.stringValue;
	
	[self loginWithCredentials:credentials];
}

- (IBAction)reloginPressed:(NSButton *)sender {
	
	P4Credentials *credentials = [[P4Credentials alloc] init];
	credentials.password = reloginPasswordField.stringValue;

	sender.enabled = NO;
	
	[self setPresentedView:loadingView animated:YES];
	[loadingIndicator startAnimation:nil];
	[loadingLabel setStringValue:@"Reconnecting..."];

	[[P4Workspace sharedInstance]
	 reloginWithCredentials:credentials
	 response:^(P4Operation *operation, NSArray *response) {
		
		 sender.enabled = YES;
		 [loadingIndicator stopAnimation:nil];

		 if (operation.errors) {
			 [self setPresentedView:reloginView animated:YES];
			 
			 PSLog(@"Failed to reconnect");
			 NSError *error = operation.error;
			 ErrorPanelController *alert = [[ErrorPanelController alloc] init];
			 [alert setTitle:@"\nFailed to reconnect"];
			 [alert setMessage:error.localizedDescription];
			 [alert presentWithMainWindow];
			 
		 } else {
			 NSString *lastWorkspace = [P4Defaults sharedInstance].workspace;
			 if (![P4Workspace sharedInstance].workspace && lastWorkspace) {
				 [self loadWorkspace:lastWorkspace];
			 } else {
				 if ([delegate respondsToSelector:@selector(loginPanelDidRelogin)])
					 [delegate loginPanelDidRelogin];
			 }
		 }
	 }];
}

- (IBAction)disconnectPressed:(id)sender {
	
	[self setPresentedView:loadingView animated:YES];
	[loadingIndicator startAnimation:nil];
	[loadingLabel setStringValue:@"Logging out..."];

	[[P4Workspace sharedInstance] logout:^(P4Operation *operation, NSArray *response) {
		[loadingIndicator stopAnimation:nil];
		[self setPresentedView:loginView animated:YES];
	}];

}

- (IBAction)ssoCheckboxPressed:(id)sender {
	allowsSSOLogin = ssoCheckbox.state == NSOnState;
}

@end
